/*
drawmap - create simple images from PoP2 level (.plv) files
Copyright (C) 2006  Hubai Tamas
created:       2006-09-21
last modified: 2006-09-23

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/*
This program uses the ImageMagick library. Make sure you have it
installed, and when linking, add -lWand.
*/

#include <stdio.h>
#include <stdlib.h>
#include <wand/magick-wand.h>

#define ThrowWandException(wand) \
{ \
  char \
    *description; \
 \
  ExceptionType \
    severity; \
 \
  description=MagickGetException(wand,&severity); \
  (void) fprintf(stderr,"%s %s %ld %s\n",GetMagickModule(),description); \
  description=(char *) MagickRelinquishMemory(description); \
  exit(-1); \
}
#define check(ret) \
if (ret == MagickFalse) \
  ThrowWandException(mw)

struct plvhead1 {
  char magic[7];
  char popversion;
  char plvversion;
  char lvlnum;
} head1;

struct plvhead2 {
  int numfields;
  int lvlsize;
} head2;

char *data;

// pointers & information about maps
char *nbh[32];
char l1[32];
char ln[32];
char c1[32];
char cn[32];
char active[33];

// which pop version
char popv;

inline int wall(int room, int tile) {
  // room number: 1..32, tile number: 0..29
  return (room-1)*30+tile;
}

/*inline int background(int room, int tile, int byte) {
  if(popv==1) {
    return 720+(room-1)*30+tile+byte;
  } else {
    return 960+(room-1)*30*4+tile*4+byte;
  }
}*/

inline int link(int room, int dir) {
  // left 0, right 1, up 2, down 3
  if(popv==1) {
    return 1952+4*(room-1)+dir;
  } else {
    return 6080+4*(room-1)+dir;
  }
}

inline int startroom() {
  if(popv==1) {
    return 2112;
  } else {
    return 6240;
  }
}

void writeval(int m, int l, int c, char r) {
  char pr;
  if(r==0) return;
  if((l<0)||(l>70)||(c<0)||(c>70)) {
    fprintf(stderr,"edge of map\n"); exit(-1);
  }
  pr=nbh[m][l*71+c];
  if(pr==0) {
    if(active[r]!=1) {
      fprintf(stderr,"stange map geometry\n"); exit(-1);
    }
    nbh[m][l*71+c]=r;
    if(l<l1[m]) l1[m]=l;
    if(l>ln[m]) ln[m]=l;
    if(c<c1[m]) c1[m]=c;
    if(c>cn[m]) cn[m]=c;
  } else if(pr!=r) {
    fprintf(stderr,"map violates symmetry rules [1]\n"); exit(-1);
  }
}

void checkback(int a, int d, int r) {
  if(a==0) return;
  if(data[link((int)a,d)]!=r) {
    fprintf(stderr,"map violates symmetry rules [2]\n"); exit(-1);
  }
}

void expand(int m, int l, int c) {
  char r; char a;
  if((l<0)||(l>70)||(c<0)||(c>70)) {
    fprintf(stderr,"edge of map\n"); exit(-1);
  }
  r=nbh[m][l*71+c];
  if(r==0) return;
  if(active[r]!=1) return;
  active[r]=0;
  a=data[link((int)r,0)]; // left
  checkback(a,1,r);
  writeval(m,l,c-1,a);
  expand(m,l,c-1);
  a=data[link((int)r,1)]; // right
  checkback(a,0,r);
  writeval(m,l,c+1,a);
  expand(m,l,c+1);
  a=data[link((int)r,2)]; // up
  checkback(a,3,r);
  writeval(m,l-1,c,a);
  expand(m,l-1,c);
  a=data[link((int)r,3)]; // down
  checkback(a,2,r);
  writeval(m,l+1,c,a);
  expand(m,l+1,c);
}

void newmap(int m, char r) {
  int i;
  nbh[m]=calloc(5041,sizeof(char));
  if(!nbh[m]) {
    fprintf(stderr,"out of memory\n"); exit(-1);
  }
  nbh[m][2520]=r;
  l1[m]=35;
  ln[m]=35;
  c1[m]=35;
  cn[m]=35;
  expand(m, 35, 35);
}

void freemap(int m) {
  free(nbh[m]);
}

void pnum(char n) {
  if(n==0) {
    printf(".. ");
  } else if(n<10) {
    printf(".%hhd ",n);
  } else {
    printf("%hhd ",n);
  }
}

void printmap(int m) {
  int l; int c; char v;
  if((l1[m]==35)&&(ln[m]==35)&&(c1[m]==35)&&(cn[m]==35)) {
    // isolated room
    active[nbh[m][2520]]=1;
    return;
  }
  printf("[map component %d]\n",m);
  for(l=l1[m];l<=ln[m];l++) {
    for(c=c1[m];c<=cn[m];c++) {
      pnum(nbh[m][l*71+c]);
    }
    printf("\n");
  }
  printf("\n");
}

void genmaps() {
  int i; int m=0; int c=0;
  for(i=1;i<=32;i++) {
    active[i]=1;
  }
  i=data[startroom()];
  //for(i=1;i<=32;i++) {
    if(active[i]) {
      newmap(m,i);
      m++;
    }
  //}
  /*for(i=0;i<m;i++) {
    printmap(i);
  }
  printf("isolated rooms: ");
  for(i=1;i<=32;i++) {
    if(active[i]) {
      if(c) printf(",");
      printf("%hhd",i);
      c++;
    }
  }
  if(!c) printf("none");
  printf("\n");*/
  /*for(i=0;i<m;i++) {
    freemap(i);
  }*/
  for(i=1;i<m;i++) {
    freemap(i);
  }
}

  MagickWand *mw; PixelWand *pw; DrawingWand *dw;
  MagickWand *ow; MagickWand *tw; MagickWand *rw;

void drawroom(int r) {
  // draws the room in r to rw (room wand)
  // uses pw, dw, tw (temp wand), ow (overlay wand)
  // these (including rw) must be preinited
  int i; int j; int si; int di; char buf[8];

  if(r==0) {
    check(PixelSetColor(pw,"grey"));
  } else {
    check(PixelSetColor(pw,"white"));
  }
  check(MagickNewImage(rw,170,99,pw));
  
  ClearDrawingWand(dw);
  DrawSetStrokeWidth(dw,1);

  if(r!=0) {
    printf("drawing room %d...\n", r);

    // place the objects
    for(di=0;di<30;di++) {
      si=data[wall(r,di)];
      if(popv==1) {
        si &= 0x1f;
      } else {
        if(si>0x2f) si=0x2f;
      }
      tw=MagickGetImageRegion(ow,16,32,1+(si%8)*17,1+(si/8)*33);
      MagickCompositeImage(rw,tw,OverCompositeOp,1+(di%10)*17,1+(di/10)*33);
    }
    
    // draw the grid
    check(PixelSetColor(pw,"#606060"));
    DrawSetStrokeColor(dw,pw);
    for(i=0;i<3;i++) {
      DrawLine(dw,0,i*33,169,i*33);
    }
    for(j=0;j<10;j++) {
      DrawLine(dw,j*17,0,j*17,98);
    }
  }
  // screen edge lines are to be drawn for blank (0) screens also
  check(PixelSetColor(pw,"darkred"));
  DrawSetStrokeColor(dw,pw); 
  DrawLine(dw,0,0,169,0);
  DrawLine(dw,0,0,0,98);
  
  if(r!=0) {
    // show room number
    sprintf(buf,"%d",r);

    DrawSetStrokeAlpha(dw,0.5);
    DrawSetFillAlpha(dw,0.5);
    DrawAnnotation(dw,8,16,buf);
    DrawSetStrokeAlpha(dw,1);
    DrawSetFillAlpha(dw,1);
  }
  
  check(MagickDrawImage(rw,dw));
}

int main(int argc, char **argv) {
  FILE *f; int r; int t; char w; char cks;
  char b1; char b2; char b3; char b4;
  int l; int c; int width; int height; int x; int y;
  int xs; int xt; int ys; int yt;
  if(argc!=2) {
    fprintf(stdout,"filename needed\n"); exit(-1);
  }
  f=fopen(argv[1],"rb");
  if(!f) {
    fprintf(stdout,"unable to open file\n"); exit(-1);
  }
  fread(&head1, 10, 1, f);
  fread(&head2, 8, 1, f);
  fread(&cks, 1, 1, f); // read the checksum byte
  data=malloc(head2.lvlsize);
  if(!data) {
    fprintf(stdout,"out of memory\n"); exit(-1);
  }
  fread(data, head2.lvlsize, 1, f);

  popv=head1.popversion;

  genmaps();
  xs=c1[0];
  xt=cn[0];
  ys=l1[0];
  yt=ln[0];

  MagickWandGenesis();
  mw=NewMagickWand();
  pw=NewPixelWand();
  dw=NewDrawingWand();
  ow=NewMagickWand();
  if(popv==1) {
    check(MagickReadImage(ow,"tiles1.png"));
  } else {
    check(MagickReadImage(ow,"tiles2.png")); 
  }
  tw=NewMagickWand();
  rw=NewMagickWand();

  check(PixelSetColor(pw,"white"));
  width=(xt-xs+1)*170+1;
  height=(yt-ys+1)*99+1;
  check(MagickNewImage(mw,width,height,pw));

  for(l=ys;l<=yt;l++) {
    y=(l-ys)*99;
    for(c=xs;c<=xt;c++) {
      r=nbh[0][l*71+c];
      x=(c-xs)*170;
      drawroom(r);
      MagickCompositeImage(mw,rw,OverCompositeOp,x,y);
    }
  }

  ClearDrawingWand(dw);
  check(PixelSetColor(pw,"darkred"));
  DrawSetStrokeColor(dw,pw); 
  DrawSetStrokeAlpha(dw,1);
  DrawSetFillAlpha(dw,1);
  DrawLine(dw,width-1,0,width-1,height-1);
  DrawLine(dw,0,height-1,width-1,height-1);
  
  check(MagickDrawImage(mw,dw));
  
  check(MagickWriteImage(mw,"drawmap.png"));
  mw=DestroyMagickWand(mw);
  pw=DestroyPixelWand(pw);
  dw=DestroyDrawingWand(dw);
  ow=DestroyMagickWand(ow);
  tw=DestroyMagickWand(tw);
  MagickWandTerminus();
  
  return 0;
}
